package cn.ibizlab.core.extensions.service;

import cn.ibizlab.core.analysis.domain.DADimension;
import cn.ibizlab.core.analysis.service.IDABuildService;
import cn.ibizlab.core.extensions.cql.ExecResult;
import cn.ibizlab.core.extensions.domain.BaseRequest;
import cn.ibizlab.core.extensions.domain.ResultsMQMsg;
import cn.ibizlab.core.extensions.util.DefaultMQProducerService;
import cn.ibizlab.core.lite.extensions.domain.ModelObj;
import cn.ibizlab.core.rule.domain.ExecLog;
import cn.ibizlab.core.rule.service.IExecLogService;
import cn.ibizlab.util.dict.CodeItem;
import cn.ibizlab.util.helper.CachedBeanCopier;
import cn.ibizlab.util.helper.DataObject;
import lombok.extern.slf4j.Slf4j;
import org.kie.api.runtime.KieSession;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Lazy;
import org.springframework.stereotype.Service;
import org.springframework.util.ObjectUtils;
import org.springframework.util.StringUtils;

import java.math.BigDecimal;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

@Slf4j
@Service("AnalyseEntityServiceImpl")
public class AnalyseEntityServiceImpl extends BaseEntityServiceImpl {
    public static final String Tag_SaveResults = "SaveResults";

    @Autowired
    private DefaultMQProducerService defaultMQProducerService;

    @Autowired
    @Lazy
    private cn.ibizlab.util.client.IBZDictFeignClient ibzDictFeignClient;

    @Override
    public void saveResult(ModelObj param, String RULEID, String RULECODE, String RULENAME, String RU_EXECRESULTNAME, Object BUSINESSCAT, Integer RETVALUE,
                           Object KEYVALUEFIELD, Object DOMAINSFIELD, Object DIMFIELD, Object METRICFIELD, Object TIMEFIELD, Object EXT1FIELD, Object EXT2FIELD)
    {
        if(0==RETVALUE)
            return;

        List<ExecResult> saveResults = new ArrayList<>();
        ExecResult result=new ExecResult();
        result.setRuleid(RULEID);
        result.setRulename(RULENAME);
        result.setRuexecresultname(RU_EXECRESULTNAME);
        result.setRetvalue(RETVALUE);
        result.setKeyvaluefield(DataObject.getStringValue(valueOf(KEYVALUEFIELD),param.getRowKey()));

        //result.setSystemId(DataObject.getStringValue(param.get("systemid"),null));
        result.setBusinesscat(DataObject.getStringValue(valueOf(BUSINESSCAT),null));
        if(METRICFIELD!=null) {
            Object val = METRICFIELD;
            if ( (val instanceof BigDecimal))
                result.setMetricfield(((BigDecimal) val).doubleValue());
            else if ( (val instanceof Double))
                result.setMetricfield(new BigDecimal((Double) val).doubleValue());
            else if ( (val instanceof Float))
                result.setMetricfield(new BigDecimal(((Float) val).doubleValue()).doubleValue());
            else if ( (val instanceof Integer))
                result.setMetricfield(new BigDecimal((int)val).doubleValue());
            else
                result.setMetricfield(1d);
        }
        else
            result.setMetricfield(1d);
        Object timefield=valueOf(TIMEFIELD);
        if(timefield!=null&&timefield instanceof Timestamp)
            result.setTimefield(ExecResult.time2LocalDate(((Timestamp)TIMEFIELD)));
        if(EXT1FIELD!=null) {
            String ext1=valuesOf(EXT1FIELD);
            if(ext1!=null&&ext1.length()>330)
                ext1=ext1.substring(0,330)+"...";
            result.setExt1field(ext1);
        }
        if(EXT2FIELD!=null) {
            String ext2=valuesOf(EXT2FIELD);
            if(ext2!=null&&ext2.length()>330)
                ext2=ext2.substring(0,330)+"...";
            result.setExt2field(ext2);
        }

        if(DOMAINSFIELD!=null)
        {
            result.setDomainsfield(DataObject.getStringValue(valueOf(DOMAINSFIELD), ""));
        }

        if((!StringUtils.isEmpty(result.getRuleid()))&&(!StringUtils.isEmpty(result.getDomainsfield()))&&(!StringUtils.isEmpty(result.getKeyvaluefield()))&&(!ObjectUtils.isEmpty(result.getTimefield()))) {
            result.setUpdatedate(ExecResult.nowLocalDate());

            List<DADimension> dims=this.getDims(result.getBusinesscat());
            if(dims!=null) {
                ArrayList<ExecResult> results = new ArrayList<>();
                for (DADimension dim : dims) {
                    Object codevalue=param.$(dim.getField());
                    if(!ObjectUtils.isEmpty(codevalue))
                    {
                        String dictname=dim.getDict();
                        Object val=codevalue;

                        if(!StringUtils.isEmpty(dictname))
                        {
                            //维度指定代码表时，向上同时为每一个父节点添加一条数据
                            while(true)
                            {
                                CodeItem code=ibzDictFeignClient.getCodeList(dictname).findCodeItem(val);
                                if(code!=null)
                                {
                                    ExecResult result2 = new ExecResult();
                                    CachedBeanCopier.copy(result, result2);
                                    result2.setDimfield(valuesOf(code.getValue()));
                                    saveResults.add(result2);
                                    if(StringUtils.isEmpty(code.getParent()))
                                    {
                                        break;
                                    }
                                    else
                                    {
                                        val=code.getParent();
                                    }
                                }
                                else
                                    break;
                            }
                        }else{
                            //维度未指定代码表时，只保存一条数据
                            result.setDimfield(valuesOf(codevalue));
                            saveResults.add(result);
                        }
                    }
                }
            }

            if(saveResults.size() > 0){
                param.set(Tag_SaveResults, saveResults);
            }
        }

    }

    private Hashtable<String, List<DADimension>> dimset=new Hashtable<>();
    private Object dimslock=new Object();

    @Autowired
    private IDABuildService daBuildExService;

    private List<DADimension>  getDims(String id)
    {
        synchronized(dimslock)
        {
            if(!dimset.containsKey(id))
                dimset.put(id, daBuildExService.get(id).getDadimension());
            return dimset.get(id);
        }
    }

    @Autowired
    protected IExecLogService ruExecLogService;
    public ExecLog processAll(BaseRequest msg)
    {

        ExecLog execlog=new ExecLog();
        execlog.setId(msg.getId());
        execlog.setRunBody(msg.toString());

        StringBuilder strErrorInfo=new StringBuilder();
        int count = 0;
        int errorcount = 0;

        String strResId=initRule(msg.getRules(),msg.getBatch());
        if (StringUtils.isEmpty(strResId))
        {
            strErrorInfo.append("准备rule文件错误");
            strErrorInfo.append("\r\n");
            execlog.setRetCode(2);
            execlog.setCnt(count);
            execlog.setSucc(errorcount);
            execlog.setRunResult(strErrorInfo.toString());
            ruExecLogService.update(execlog);
            return execlog;
        }

        try
        {


            for (ModelObj modelObj : msg.getDatas())
            {
                try
                {

                    count++;
                    modelObj.set("BATCH",msg.getBatch());
                    modelObj.set(Tag_EngineId, msg.getEngineId());
                    modelObj.set("systemid",msg.getSystemid());
                    if(!StringUtils.isEmpty(msg.getResultDataSource())){
                        modelObj.set(RuleEngineExService.Setting_ResultDataSource, msg.getResultDataSource());
                    }
                    if(!StringUtils.isEmpty(msg.getResultTableName())){
                        modelObj.set(RuleEngineExService.Setting_ResultTableName, msg.getResultTableName());
                    }
                    processRule(modelObj, modelObj.getRowKey(),strResId);
                    ResultsMQMsg resultsMQMsg = new ResultsMQMsg();
                    resultsMQMsg.setKeyValueField(modelObj.getRowKey());
                    resultsMQMsg.setRules(msg.getRuleIds());
                    Object saveResults = modelObj.get(Tag_SaveResults);
                    if(saveResults != null && saveResults instanceof List){
                        resultsMQMsg.setDatas((List<ExecResult>)saveResults);
//                        defaultMQProducerService.sendToMQ(resultsMQMsg, "default");
                    }
                    defaultMQProducerService.sendToMQ(resultsMQMsg, "default");
                }
                catch (Exception e)
                {
                    errorcount++;
                    log.error("Service错误，加载数据详细信息：" + modelObj.getRowKey() + "," + e.getMessage());
                    if (strErrorInfo.toString().length() <= 5000)
                    {
                        strErrorInfo.append("加载数据详细信息：" + modelObj.getRowKey() + ",错误：" + e.getMessage());
                        strErrorInfo.append("\r\n");
                        if (e.getCause() != null && e.getCause().getMessage() != null)
                        {
                            strErrorInfo.append(e.getCause().getMessage());
                            strErrorInfo.append("\r\n");
                        }

                    }

                }

            }






        }
        catch(Exception e)
        {
            log.error("Service读取BaseRequest错误:" + e.toString());
            strErrorInfo.append("读取BaseRequest错误:" + e.getMessage());
            strErrorInfo.append("\r\n");
            if (e.getCause() != null && e.getCause().getMessage() != null)
            {
                strErrorInfo.append(e.getCause().getMessage());
                strErrorInfo.append("\r\n");
            }


            execlog.setRetCode(1);
            execlog.setCnt(count);
            if (errorcount == 0)
                execlog.setSucc(0);
            else
                execlog.setSucc(count-errorcount);
            execlog.setRunResult(strErrorInfo.toString());
            ruExecLogService.update(execlog);
            return execlog;
        }
        finally
        {
            if(setkieSession.containsKey(strResId))
            {
                KieSession kieSession=setkieSession.get(strResId);
                try
                {
                    kieSession.dispose();
                }
                catch (Exception ex){}

                setkieSession.remove(strResId);
                setRuleCount.remove(strResId);
            }

        }

        log.info(new StringBuilder().append("Service成功，process successfully:").append(msg.toString()).toString());
        strErrorInfo.append("Service成功，process successfully\r\n");





        execlog.setCnt(count);
        execlog.setSucc(count-errorcount);
        if(execlog.getCnt()==0)
            execlog.setRetCode(3);
        else
            execlog.setRetCode(0);
        execlog.setRunResult(strErrorInfo.toString());
        this.ruExecLogService.update(execlog);
        return execlog;
    }
}
